;;;;
;;;; p r o o f . s c m          -- A very simple and stupid reflective
;;;;                               proof checking evaluator for
;;;;                               a pure subset of Scheme
;;;;
;;;; Copyright C 1997 Far Rideau under the GPL
;;;;
;;;;           Author: Far Rideau [rideau@unice.fr]
;;;;

;;;; See README for description

(provide 'proof)
(require 'macrolib)	; a few macros...

;(display "I'm here")(newline)(exit) ; dichotomie-boot

(require 'lib)		; library of general-purpose routines
(require 'tags)		; unique tags for user-defined types
(require 'debug)	; debugging...
(require 'pmstruct)	; poor man's structures
(require 'theenv)	; global environment -- yuck
(require 'pattern)	; pattern matching

;;;; Just stuff...

(*test-off*)
(*test* ((make-ref 1)
	 '#((type ref $) 1))
	((ref:post-inc (make-ref 1))
	1)
)

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Catching errors
;;;;
(define (soft-error . foo) (apply error foo))
(define (soft-assert . foo) (apply $assert foo))


(set! **debug** #f)

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Grammatical constructors for a pure subset of Scheme
;;;;


(define-macro% ($apply sym . args)
  "apply a function from THE-ENV"
  `(apply ($ ,sym) ,@args))
(define-macro% (register-pattern pattern)
  "register a basic grammatical form"
  `(add-struct-definitions! THE-ENV ',pattern))

(register-pattern (lambda formals body))		; lambda
(register-pattern (varref sym num))			; lambda
(register-pattern (syntax function params))		; syntax
(register-pattern (quote obj))				; quote
(register-pattern (appl head args))			; appl
(register-pattern (syntax formals body))		; syntax
(register-pattern (T<% lhs rhs))			; T<%
(register-pattern (either left right))			; either
(register-pattern (both left right))			; both
(register-pattern (special constructor matcher))	; special
(register-pattern (cons head tail))			; pure cons

(*test-off*)
(*test*
 (THE-ENV
  THE-ENV)
 (($ make-lambda '(x y) 'x)
	 '#((lambda $) (x y) x))
 (($ quote? ($ make-quote 'x))
         '#t)
)





;;;; deconstruction...
(define (constr:deconstr x)
  (cond
   ((vectree? x)
      (vector->list x))
   ((pair? x)
      (list cons-tag (car x) (cdr x)))
   (else #f)))
(define (constr:constr k x)
  (if (eq? k cons-tag)
      (apply cons x)
      (apply vector k x)))
(define (constr=? k1 k2) (eq? k1 k2))
(define (constr? x)
  (or (vectree? x) (pair? x)))

(*test-off*)
(*test* ((constr:deconstr #t)
	 #f)
	((constr:deconstr ($ make-appl 'f '(x)))
	 '((appl $) f (x)))
)


;;;; Builtin objects
(define (constant? x)
  (or (boolean? x)
      (null? x)
      (char? x)
      (number? x)
      (eof-object? x)))
(define (builtin? x)
  (or (constant? x)
      (string? x)
      (procedure? x)))


;;;; Primitives
(define *primitives* (env:new))
(define (register-primitive sym val)
  (env:add-definition! *primitives* sym val))
(define (primitive? x)
  (env:bound? *primitives* x))

(define-macro% (define-builtin-primitive keyword)
  `(register-primitive ',keyword ,keyword))

(map-macro (begin) define-builtin-primitive
     read write display newline
     eq?
;     memq assq
     equal?
;     member assoc
;     eqv? memv assv
     car cdr cons pair?
     null? list append
;     caar caaar caaaar cdaaar
;          cdaar cadaar cddaar
;     cdar cadar caadar cdadar
;          cddar caddar cdddar
;     cadr caadr caaadr cdaadr
;          cdadr cadadr cddadr
;     cddr caddr caaddr cdaddr
;          cdddr cadddr cddddr
     + -
;     * /
;     = < <= > >=
     exit
)

(define-macro% (define-primitive keyword)
  `(define ,keyword (register-primitive ',keyword (make-tag ',keyword))))

(define-primitive ==>>)		; does $1 evaluate to $2 ?
(define-primitive ==>)		; does $1 reduce to $2 ?
(define-primitive ==)		; is $1 observationally equal to $2 ?
;(define-primitive allv?)	; does $1 stand for all values ?
(define-primitive normal?)	; is $1 in normal form ?
(define-primitive existv?)	; does there exist a value satisfying $1 ?
(define-primitive either)	; either $1 or $2
(define-primitive pi)		; a pi-channel


;;;; Keywords and symbols
(define *keywords* (env:new))
(define (register-keyword sym val)
  (env:add-definition! *keywords* sym val))
(define (keyword? x)
  #f)
;  (env:bound? *keywords* x))
(define (varname? x)
  (and (symbol? x) (not (keyword? x))))


;;;; Different type of objects...
(define (self-eval? x)
  (or (builtin? x)
      ($ lambda? x)))

(define (self-reduce? x)
  (or (self-eval? x)
      ($ quote? x)))


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Parser
;;;;

; special forms are not allowed as values
(define (ensure-not-special foo)
  (if ($ special? foo)
      (error "special result not allowed as value" foo)
      foo))

; parsing lambda patterns
(define (ensure-lambda-pattern baz)
  (if (lambda-pattern->vars baz)
      baz
      (error "lambda pattern required" baz)))

(define (lambda-pattern->vars-0 pat kvars)
  (cond
   ((null? pat)
    kvars)
   ((varname? pat)
    (cons pat kvars))
   ((and (pair? pat)
	 (varname? (car pat)))
    (lambda-pattern->vars-0 (cdr pat) (cons (car pat) kvars)))
   (else #f)))

(define (lambda-pattern->vars formals)
  (lambda-pattern->vars-0 formals '()))

(define (parse-formals formals)
;  (letrec ((fun
;	     (lambda (rem n pat)
;	       (cond
;		((null? pat) (list rem n #f))
;		((varname? pat) (list rem n #t))
;		((pair? pat) (fun rem (+ n 1) (cdr pat)))))))
; (fun formals 0 formals))
  formals)

(define (vars->frame vars)
  (make-frame (map (lambda (var) (make-binding var #t)) vars)))

(define (parse-lambda env formals body)
  (let ((vars (lambda-pattern->vars formals)))
    (if (not vars) (error "lambda pattern required" formals))
    (let*
	((frame  (vars->frame vars))
	 (newenv (env:cons frame env)))
    ($ make-lambda
      (parse-formals formals)
      (parse-in-env newenv body)))))

(define (parse-symbol env sym)
  (env:try-get-with-depth
     env sym
     (lambda (val n)
       (if (eq? val #t)
	   ($ make-varref sym n)
	   val))
     (lambda (n) sym)))

(define (parse-args env args)
  (map (lambda (arg) (ensure-not-special (parse-in-env env arg))) args))

(define (parse-quote env obj)
  ($ make-quote obj))

(define (parse-quote-parsed env obj)
  ($ make-quote (parse-in-env env obj)))

(define (parse-special form env foo)
  (apply ($ special->constructor form) env foo))

(define (parse-in-env env foo)
  (cond
   ((self-eval? foo) ($ make-quote foo))
   ((varname? foo)
    (parse-symbol env foo))
   ((list? foo)
    (let ((form (parse-in-env env (car foo)))
	  (args (cdr foo)))
    (if ($ special? form)
	(parse-special form env args)
	($ make-appl form (parse-args env args)))))
   (else
    (error "Cannot parse" foo))))


;;;; Special forms
(define *special-forms* (env:new))
(define (register-special-form sym val)
  (env:add-definition! *special-forms* sym ($ make-special val #f)))

(register-special-form 'quote		parse-quote)
(register-special-form 'lambda		parse-lambda)
;(register-special-form 'quote-parsed	parse-quote)


;;;; Initial environment
(define init-env (env:merge *primitives* *special-forms*))
(define (parse foo) (parse-in-env init-env foo))


(*test-off*)
(*test*
  (init-env init-env)
  ((env:try-get-with-depth
     *primitives* '+
     (lambda (val n)
       (if (eq? val #t)
	   ($ make-varref '+ n)
	   val))
     (lambda (n) '+))
  +)
  ((env:try-get-with-depth
     init-env '+
     (lambda (val n)
       (if (eq? val #t)
	   ($ make-varref '+ n)
	   val))
     (lambda (n) '+))
  +)
  ((parse '+) +)
)

;;;; Let's test the parser...
(*test-off*)
(*test*
        (($ special-tag)
	 '(special $))
 	((parse #t)
	 '#((quote $) #t))
 	((parse '+)
	 +)
	((parse '(lambda (foo) foo))
	 '#((lambda $) (foo) #((varref $) foo 0)))
	((parse '(lambda (foo) (foo bar)))
	 '#((lambda $) (foo)
		   #((appl $) #((varref $) foo 0) (bar))))
	((parse ''(foo))
	 '#((quote $) (foo)))
	((parse '(lambda (foo) (lambda (bar) foo)))
	 '#((lambda $) (foo) #((lambda $) (bar) #((varref $) foo 1))))
	;((parse '(lambda (foo . 1) (foo bar))) foo) ; error expected!
)

(define *t*id (parse '(lambda (foo) foo)))
(define *t*bar (parse ''bar))
(define *t*e1 (parse `(,+ 1 2)))
(define *t*e2 (parse `(,+ 1 2)))

(*test*
 ((begin (format #t "~s\n~s\n~s\n~s\n" *t*id *t*bar *t*e1 *t*e2) #f)
  #f)
 (*t*id
  '#((lambda $) (foo) #((varref $) foo 0)))
)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Beta-reduction
;;;;
(define (do-shift-ref shift depth ref)
  "shift by SHIFT reference REF in a DEPTH deep subobject"
  (let* ((s ($ varref->sym ref))
	 (n ($ varref->num ref)))
    (if (< n depth) ref
	($ make-varref s (+ n shift)))))

(define (shift-ref shift depth)
  "shift by SHIFT all references in a DEPTH deep subobject"
  (lambda (obj)
    (cond
     (($ appl? obj)
      ($ make-appl
	 ((shift-ref shift depth) ($ appl->head obj))
	 (map (shift-ref shift depth) ($ appl->args obj))))
     (($ lambda? obj)
         (let* ((formals ($ lambda->formals obj))
		(vars    (lambda-pattern->vars formals))
		(body    ($ lambda->body obj)))
         ($ make-lambda formals
	    ((shift-ref shift (+ 1 depth)) body))))
     ((self-reduce? obj) obj)
     (($ varref? obj) (do-shift-ref shift depth obj))
     (else (error "shift-ref" shift depth obj)))))


(define (do-subst-ref depth frame ref)
  "substitute arguments from FRAME DEPTH frames away"
  (let* ((num ($ varref->num ref))
	 (sym ($ varref->sym ref))
	 (shifted (- num depth)))
;  (dbgfmt "DO-SUBST-REF ref: ~s orig: ~s shift: ~s shifted: ~s\n"
;	  sym num depth shifted)
    (cond
      ((< shifted 0) ref)
      ((= shifted 0)
        (let ((val (frame:get frame sym)))
;	  (dbgfmt "SHIFT-REF ~s\n" val)
	  ((shift-ref depth 0) val)))
      (else ; (> shifted 0)
        (error "DO-SUBST-REF" depth frame ref)
	; ref should already have been replaced
       	($ make-varref ($ varref->sym ref) (- num 1))))))

(define (do-subst-lambda depth frame lam)
"substitute args in a lambda sub-object protected
 by DEPTH levels or abstraction frames"
  (let* ((formals ($ lambda->formals lam))
;	 (vars    (lambda-pattern->vars formals))
	 (body    ($ lambda->body lam)))
    ($ make-lambda formals
		 ((do-subst-obj (+ 1 depth) frame) body))))

(define (do-subst-obj depth frame)
"substitute frame in a sub-object protected
 by DEPTH levels or abstraction frames"
  (lambda (obj)
;    (dbgfmt "Dsubstobj ~a [[ ~a ]] ~a\n" depth frame obj)
    (cond
     (($ appl? obj)
;         (dbgfmt "DSO-appl\n")
         ($ make-appl
	    ((do-subst-obj depth frame) ($ appl->head obj))
	    (map (do-subst-obj depth frame) ($ appl->args obj))))
     (($ lambda? obj)
;         (dbgfmt "DSO-lam\n")
         (do-subst-lambda depth frame obj))
     (($ varref? obj)
;         (dbgfmt "DSO-ref\n")
	 (do-subst-ref depth frame obj))
     ((self-reduce? obj)
;         (dbgfmt "DSO-sr\n")
	 obj))))

(define (subst-obj frame obj)
  "substitute bindings in obj"
  ((do-subst-obj 0 frame) obj))

(define (match-args-into-frame formals args)
  "build a frame with formal arguments bound to actual arguments"
  (let ((frame (try-match formals args)))
    (if frame frame (soft-error "argument mismatch" formals args))))

(define (do-beta-reduce lam args)
   "do apply beta-reduction to a lambda expression LAM
applied to arguments args"
   (soft-assert ($ lambda?) lam "not a lambda to beta-reduce")
   (let ((formals ($ lambda->formals lam))
         (body    ($ lambda->body lam)))
     (dbgfmt "FORMALS: ~s\nARGS: ~s\nBODY: ~s\n" formals args body)
     (subst-obj (match-args-into-frame formals args) body)))

;(traceon do-beta-reduce)
;(traceon subst-obj)
;(traceon do-subst-lambda)
;(traceon do-subst-ref)
;(traceon shift-ref)
;(traceon do-shift-ref)

(*test-off*)
(define *t*twu (parse '(lambda (add one S O) ((add one one) S O))))
(define *t*add (parse '(lambda (n m) (lambda (S O) (n S (m S O))))))
(define *t*one (parse '(lambda (S O) (S O))))
(define *t*S   (parse '(lambda (x) (+ x 1))))
(define *t*O   (parse '0))
(define *t*two ($ make-appl *t*twu (list *t*add *t*one *t*S *t*O)))
;(dbgfmt "~s\n~s\n~s\n~s\n~s\n~s\n" *t*twu *t*add *t*one *t*S *t*O *t*two)

(*test* ((subst-obj (make-frame (list (make-binding 'y 2)))
		    ($ make-lambda '(x) ($ make-varref 'y 1)))
	 '#((lambda $) (x) 2))
        ((do-beta-reduce *t*id (list *t*bar))
	 '#((quote $) bar))
	(*t*twu
	 '#((lambda $) (add one s o)
			   #((appl $)
			     #((appl $) #((varref $) add 0)
					(#((varref $) one 0)
					 #((varref $) one 0)))
			    (#((varref $) s 0) #((varref $) o 0)))))
	(*t*one
	 '#((lambda $) (s o)
	    #((appl $) #((varref $) s 0) (#((varref $) o 0)))))
)
;(exit)

;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Straigthforward evaluation of lambdas
;;;;
(define (repr->value x)
  (cond
   (($ quote? x) ($ quote->obj x))
   ((self-eval? x) x)
   (else (soft-error "not a quoted value!" x))))
(define (expose-primitive prim)
  (lambda (args)
;    (dbgfmt "expose-prim ~a ~a\n" prim args)
    ($ make-quote (apply prim (map repr->value args)))))
(define (compiled x)
  (dbgfmt "compiled: ~a\n" x)
  (cond
   ((procedure? x) x)
   ((and ($ quote? x) (procedure? ($ quote->obj x)))
    ($ quote->obj x))
   (else #f)))
(define (apply-compiled fun args)
  ((expose-primitive fun) args))

(define (simple-reduce foo)
  (cond
    ((self-reduce? foo) foo)
;    (($ quote? foo) ($ quote->obj foo)) ;NO!
    (($ lambda? foo) foo)
    (($ appl? foo)
;       (dbgfmt "EVAL: ")
;       (dbgfmt "~a\n" foo)
       (let* ((head ($ appl->head foo))
	      (args ($ appl->args foo))
;	      (xxxx (dbgfmt "xxx ~s\n" args))
	      (ehead (simple-reduce head))
	      (eargs (map simple-reduce args))
	      (comp (compiled ehead)))
	 (dbgfmt "compiled=> ~a\n" comp)
	 (cond
	  (comp (apply-compiled comp eargs))
	  (($ lambda? ehead)
;	       (dbgfmt "azeaze\n")
	       (simple-reduce (do-beta-reduce ehead eargs)))
	  (else
	   (soft-error "cannot reduce any further [" ehead eargs "]\n")))))
    ((varname? foo) (error "unbound variable" foo))
    (($ varref? foo) (error "uh? asked to eval" foo))
    (else (error "unrecognized form" foo))))

(define (simple-eval foo)
  (repr->value (simple-reduce foo)))

;(traceon simple-eval)

(*test-off*)
(*test* ((simple-eval *t*e1)
	 3)
	((simple-eval ($ make-appl *t*S (list *t*O)))
	 1)
	((simple-eval *t*two)
	 2)
	((simple-eval (parse '''a))
	 ''a)
)
;(exit)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; a read-eval-print loop for the former evaluator
;;;;

(define (simple-repl)
  (define (input)
    (let ((sexp (read)))
      (format #t ">> ~a\n" sexp)
      sexp))

  (define (output val)
    (format #t "~a\n" val))

  (define (loop cont form)
    (if (eof-object? form)
	(format #t "Be seeing you!\n")
	(begin
	  (output (simple-eval (parse form)))
	  (loop cont (input)))))

  (loop (lambda (x) (write x)) (input)))
;  (call/cc (lambda (cont) (loop cont (input)))))

;(simple-repl)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Simple pattern-matching
;;;;


(*test-off*)
(*test* ((try-match-expr (frame:new) ($ make-quote #t) #t)
	 (frame:new))
	((try-match-expr (frame:new) 'a #t)
	 (make-frame (list (make-binding 'a '#t))))
)


;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; Proof steps
;;;;

(define (P:top-elim env expr)
    (or (eqv? #t expr)
	(soft-error "no match in P:top-elim")))
; top-elim's computation is implicit, by the very representation used!
; it is used only as a [?!] check

(define (P:refl-T<%-elim env expr)
    (if (try-match (parse '(T<% foo foo)) expr)
	#t
	(soft-error "no match")))
; top-elim's computation is implicit, by the very representation used!
; it is used only as a [?!] check


;(P:refl '() (*****)) ;parse '(T<% (*****


(*test-on*)
(*test*
 ((P:top-elim init-env #t)
  #t)
)



;;;;
;;;; Proof-checking evaluator
;;;;



(define (P:eval k r e)
   (*****))

;(define ())


;(*test-on*)
;(*test* ((P:top-elim #t)))
(exit)




;------>8------>8------>8------>8------>8------>8------>8------>8------>8------
;;;;
;;;; To finish, a read-eval-print loop
;;;;

;(define (simple-repl)
;  (define (input)
;    (let ((sexp (read)))
;      (format #t ">> ~a\n" sexp)
;      sexp))
;
;  (define (output val)
;    (format #t "~a\n" val))
;
;  (define (loop cont form)
;    (if (eof-object? form)
;	(format #t "Be seeing you!\n")
;	(begin
;	  (output (simple-eval (parse form)))
;	  (loop cont (input)))))
;
;  (call/cc (lambda (cont) (loop cont (input)))))
;
;(simple-repl)
